"""
Disable all active bots in a project.

1. Provide name of the project to connect to
2. Connect to the environment using data from workstation
3. Get a list of all bots in the project
4. Disable every active bot on the list
"""

from mstrio.connection import get_connection
from mstrio.project_objects.bots import list_bots

# Define project to connect to
PROJECT_NAME = 'MicroStrategy Tutorial'

# Connect to environment without providing user credentials
# Variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get a list of bots in the project
bots = list_bots(connection=conn)

# Disable every active bot
for bot in bots:
    if bot.status == 'enabled':
        bot.disable()



from enum import IntFlag
from typing import Any, TypeVar

from mstrio.api import browsing
from mstrio.connection import Connection
from mstrio.object_management.folder import Folder
from mstrio.object_management.search_enums import SearchDomain
from mstrio.object_management.search_operations import full_search
from mstrio.server.project import Project
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin, ObjectTypes
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    fetch_objects,
    get_default_args_from_func,
    get_valid_project_id,
)
from mstrio.utils.response_processors import objects as objects_processors


class ShortcutInfoFlags(IntFlag):
    DssDashboardShortcutInfoBookmark = 0b10
    DssDashboardShortcutInfoTOC = 0b01
    DssDashboardShortcutInfoDefault = 0b00


def list_shortcuts(
    connection: Connection,
    name: str | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    project: Project | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    folder: Folder | None = None,
    folder_id: str | None = None,
    folder_path: str | None = None,
    **filters,
):
    """List all shortcuts in a project.

    The project may be specified by either `project_id`, `project_name` or
        `project`. If the project is not specified in either way, the project
        from the `connection` object is used.

    Args:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`.
        name (string, optional): The search pattern for listing shortcuts.
            Supports wildcards '*' (any number of characters) and '?' (exactly
            one character).
        project_id (str, optional): ID of the project to search in.
        project_name (str, optional): Name of the project to search in.
            May be used instead of `project_id`.
        project (Project, optional): Project object specifying the project to
            search in. May be used instead of `project_id`.
        to_dictionary (bool, optional): If True, the method will return
            dictionaries with the shortcuts' properties instead of Shortcut
            objects. Defaults to False.
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        folder (Folder, optional): Folder object specifying where the search
            will be performed.
        folder_id (string, optional): ID of a folder where the search
            will be performed. Can be provided as an alternative to `folder`
            parameter.
        folder_path (str, optional): Path of the folder in which the search
            will be performed. Can be provided as an alternative to `folder`
            and `folder_id` parameters. The path has to be provided in the
            forward-slash format, beginning with project name, e.g.
            "/MicroStrategy Tutorial/Public Objects/Metrics".
        **filters: Available filter parameters: ['id', 'type', 'subtype',
            'date_created', 'date_modified', 'version', 'owner', 'ext_type',
            'view_media', 'certified_info']

    """

    if project_id is None and project is not None:
        project_id = project.id
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=True,
    )
    if folder and not folder_id:
        folder_id = folder.id
    if folder_path and folder_id:
        folder_path = None

    # No endpoint for listing shortcuts. Using full_search instead.
    objects = full_search(
        connection,
        object_types=[ObjectTypes.SHORTCUT_TYPE],
        domain=SearchDomain.PROJECT,
        project=project_id,
        name=name,
        limit=limit,
        root=folder_id,
        root_path=folder_path,
        **filters,
    )

    if to_dictionary:
        return objects

    return Shortcut.bulk_from_dict(source_list=objects, connection=connection)


class Shortcut(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """Representation of a Shortcut object. It points to another object in
    the metadata, and for Library objects it stores additional information
    related to browsing the Library.

    Attributes:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`.
        id (str): ID of the shortcut object
        name (str): Name of the shortcut
        project_id (str): ID of the project that the shortcut is in
        type (ObjectTypes): object type (SHORTCUT_TYPE)
        date_created (DateTime): creation time
        date_modified (DateTime): last modification time
        version (str): object version ID
        acg (Rights): access rights
            (See EnumDSSXMLAccessRightFlags for possible values)
        acl (list[ACE]): object access control list
        hidden (bool): Specifies whether the object is hidden
        owned_by_current_user (bool): whether the current user owns the shortcut
        target (dict): metadata of the target object related to viewing
            in Library
        target_info (dict): general metadata of the target object
        encode_html_content (bool): Flag indicating whether to encode
            HTML content
        current_page_key (str): Current page node key
        stid (int): Shortcut State ID
        current_bookmark (dict): Name and ID of the current bookmark
        prompted (bool): Flag indicating whether the target object has prompts
        datasets_cache_info_hash (str): The hash value for the Dashboard dataset
            cache info
        shortcut_info_flag (bool): flag indicating what information about
            the shortcut should be loaded
    """

    _OBJECT_TYPE = ObjectTypes.SHORTCUT_TYPE
    _API_GETTERS = {
        (
            'name',
            'id',
            'project_id',
            'owned_by_current_user',
            'target',
            'encode_html_content',
            'current_page_key',
            'shared_time',
            'last_viewed_time',
            'last_modified_time',
            'stid',
            'current_bookmark',
            'prompted',
            'datasets_cache_info_hash',
            'shortcut_info_flag',
            'dossier_version_hash',
        ): browsing.get_shortcut,
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
        ): objects_processors.get_info,
    }
    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'shortcut_info_flag': ShortcutInfoFlags}
    _API_PATCH: dict = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'comments',
            'owner',
        ): (
            objects_processors.update,
            'partial_put',
        )
    }

    def __init__(
        self,
        connection: Connection,
        id: str,
        project_id: str = None,
        project_name: str = None,
        shortcut_info_flag: (
            ShortcutInfoFlags | int
        ) = ShortcutInfoFlags.DssDashboardShortcutInfoTOC,
    ):
        """Initialize the Shortcut object and populate it with I-Server data.

        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Note:
            When `project_id` is `None` and `project_name` is `None`, then
            its value is overwritten by `project_id` from `connection` object.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`.
            id: Shortcut ID
            project_id: ID of the project that the shortcut is in
            project_name: Project name
            shortcut_info_flag: flag indicating what information about shortcut
                should be loaded
        """
        if id is None:
            raise AttributeError("Please specify 'id' parameter in the constructor.")
        else:
            project_id = get_valid_project_id(
                connection=connection,
                project_id=project_id,
                project_name=project_name,
                with_fallback=not project_name,
            )
            super().__init__(
                connection=connection,
                object_id=id,
                project_id=project_id,
                shortcut_info_flag=get_enum_val(shortcut_info_flag, ShortcutInfoFlags),
            )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._owned_by_current_user = kwargs.get('owned_by_current_user')
        self._target = kwargs.get('target')
        self._encode_html_content = kwargs.get('encode_html_content')
        self._shortcut_info_flag = (
            ShortcutInfoFlags(kwargs["shortcut_info_flag"])
            if kwargs.get("shortcut_info_flag")
            else None
        )
        self._current_page_key = kwargs.get('current_page_key')
        self._shared_time = kwargs.get('shared_time')
        self._last_viewed_time = kwargs.get('last_viewed_time')
        self._last_modified_time = kwargs.get('last_modified_time')
        self._stid = kwargs.get('stid')
        self._current_bookmark = kwargs.get('current_bookmark')
        self._prompted = kwargs.get('prompted')
        self._datasets_cache_info_hash = kwargs.get('datasets_cache_info_hash')
        self._dashboard_version_hash = kwargs.get('dossier_version_hash')

    T = TypeVar('T')

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Connection,
        to_snake_case: bool = True,
        with_missing_value: bool = False,
    ) -> T:
        """Instantiate a Shortcut from a dict source."""
        return super(Entity, cls).from_dict(
            source=source,
            connection=connection,
            to_snake_case=to_snake_case,
            with_missing_value=with_missing_value,
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ):
        """Alter the shortcut's properties.

        Args:
            name (str, optional): name of the shortcut object
            description (str, optional): description of the shortcut object
            abbreviation (str, optional): abbreviation of the shortcut object
            hidden (bool, optional): specifies whether the shortcut is hidden
            comments (str, optional): long description of the shortcut
            owner: (str, User, optional): owner of the shortcut object
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def create_shortcut(
        self,
        target_folder_id=None,
        target_folder_path=None,
        target_folder=None,
        project_id=None,
        project_name=None,
        project=None,
        to_dictionary=False,
    ):
        raise ValueError("Shortcut cannot refer to another shortcut.")

    @property
    def owned_by_current_user(self):
        return self._owned_by_current_user

    @property
    def target(self):
        return self._target

    @property
    def encode_html_content(self):
        return self._encode_html_content

    @property
    def shortcut_info_flag(self):
        return self._shortcut_info_flag

    @property
    def current_page_key(self):
        return self._current_page_key

    @property
    def shared_time(self):
        return self._shared_time

    @property
    def last_viewed_time(self):
        return self._last_viewed_time

    @property
    def last_modified_time(self):
        return self._last_modified_time

    @property
    def stid(self):
        return self._stid

    @property
    def current_bookmark(self):
        return self._current_bookmark

    @property
    def prompted(self):
        return self._prompted

    @property
    def datasets_cache_info_hash(self):
        return self._datasets_cache_info_hash

    @property
    def dashboard_version_hash(self):
        return self._dashboard_version_hash


def get_shortcuts(
    connection: Connection,
    shortcut_ids: list[str],
    project_id: str = None,
    project_name: str = None,
    shortcut_info_flag: (
        ShortcutInfoFlags | int
    ) = ShortcutInfoFlags.DssDashboardShortcutInfoDefault,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list[dict] | list[Shortcut]:
    """Retrieve information about specific published shortcuts
     in specific project.

     Specify either `project_id` or `project_name`.
     When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
         When `project_id` is `None` and `project_name` is `None`,
         then its value is overwritten by `project_id` from `connection`
         object.

     Args:
         shortcut_ids: ids of target shortcuts
         project_id: id of project that the shortcuts are in
         project_name: Project name
         shortcut_info_flag: a single ShortcutInfoFlags that describes what
           exact info are to be fetched
         to_dictionary: parameter describing output format
         limit (int): limit the number of elements returned.
             If `None` (default), all objects are returned.
     Return:
         list of dictionaries or Shortcut objects,
           depending on `to_dictionary` parameter
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    shortcuts = fetch_objects(
        connection=connection,
        api=browsing.get_shortcuts,
        dict_unpack_value="shortcuts",
        limit=limit,
        filters=filters,
        body=[{"projectId": project_id, "shortcutIds": shortcut_ids}],
        shortcut_info_flag=get_enum_val(shortcut_info_flag, ShortcutInfoFlags),
    )

    if to_dictionary:
        return shortcuts
    else:
        return [
            Shortcut.from_dict(source=short, connection=connection)
            for short in shortcuts
        ]



from enum import Enum


# A sentinel object to detect if an attribute has been returned by REST API
class _MissingType:
    pass


MISSING = _MissingType()


class ObjectTypes(Enum):
    # https://www2.microstrategy.com/producthelp/Current/ReferenceFiles/reference/com/microstrategy/webapi/EnumDSSXMLObjectTypes.html
    FILTER = 1
    TEMPLATE = 2
    REPORT_DEFINITION = 3
    METRIC = 4
    AUTO_STYLE = 6
    AGG_METRIC = 7
    FOLDER = 8
    SUBSCRIPTION_DEVICE = 9
    PROMPT = 10
    FUNCTION = 11
    ATTRIBUTE = 12
    FACT = 13
    DIMENSION = 14
    TABLE = 15
    DATAMART_REPORT = 16
    FACT_GROUP = 17
    SHORTCUT_TYPE = 18
    RESOLUTION = 19
    MONITOR = 20
    ATTRIBUTE_FORM = 21
    SCHEMA = 22
    FORMAT = 23
    CATALOG = 24
    CATALOG_DEFINITION = 25
    COLUMN = 26
    PROPERTY_SET = 28
    DBROLE = 29
    DBLOGIN = 30
    DBCONNECTION = 31
    PROJECT = 32
    SERVER_DEFINITION = 33
    USER = 34
    USERGROUP = 34
    SUBSCRIPTION_TRANSMITTER = 35
    CONFIGURATION = 36
    REQUEST = 37
    SEARCH = 39
    SEARCH_FOLDER = 40
    DATAMART = 41
    PACKAGE_DEFINITION = 42
    ROLE = 43
    SECURITY_ROLE = 44
    LOCALE = 45
    CONSOLIDATION = 47
    CONSOLIDATION_ELEMENT = 48
    SCHEDULE_EVENT = 49
    SCHEDULE_OBJECT = 50
    SCHEDULE_TRIGGER = 51
    LINK = 52
    DBTABLE = 53
    TABLE_SOURCE = 54
    DOCUMENT_DEFINITION = 55
    DRILL_MAP = 56
    DBMS = 57
    SECURITY_FILTER = 58
    PROMPT_ANSWER = 59
    PROMPT_ANSWERS = 60
    GRAPH_STYLE = 61
    BLOB = 63
    OBJECT_TAG = 65
    CHANGE_JOURNAL = 66
    SHORTCUT = 67
    SHORTCUT_TARGET = 68
    RECONCILIATION = 69
    LAYER = 70
    PALETTE = 71
    THRESHOLDS = 72
    DASHBOARD_PERSONAL_VIEW = 73
    FEATURE_FLAG = 74
    SCRIPT = 76
    CONTENT_BUNDLE = 77
    APPLICATION = 78
    TIMEZONE = 79
    RUNTIME = 80
    CALENDAR = 81
    IAM = 82
    KPI_WATCHER = 83
    DRIVER = 84
    INTERFACE_LANGUAGE = 85
    WORKFLOW = 86
    NOT_SUPPORTED = None

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value

    @classmethod
    def _missing_(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        member._name_ = f'UNDEFINED_TYPE:{value}'
        return member

    @staticmethod
    def contains(item):
        if isinstance(item, ObjectTypes):
            return True
        return item in [v.value for v in ObjectTypes.__members__.values()]

    def get_subtypes(self):
        # This takes advantage of how subtypes are defined in DSS.
        # All subtypes are defined as the corresponding type followed
        # by 8 bits of subtype ID
        return [
            subtype
            for subtype in ObjectSubTypes
            if subtype.value and subtype.value >> 8 == self.value
        ]


class ObjectSubTypes(Enum):
    # https://www2.microstrategy.com/producthelp/Current/ReferenceFiles/reference/com/microstrategy/webapi/EnumDSSXMLObjectSubTypes.html
    FILTER = 256
    CUSTOM_GROUP = 257
    TEMPLATE = 512
    REPORT_GRID = 768
    REPORT_GRAPH = 769
    REPORT_ENGINE = 770
    REPORT_TEXT = 771
    REPORT_DATAMART = 772
    REPORT_BASE = 773
    REPORT_GRID_AND_GRAPH = 774
    REPORT_NON_INTERACTIVE = 775
    OLAP_CUBE = 776
    INCREMENTAL_REFRESH_REPORT = 777
    REPORT_TRANSACTION = 778
    SUPER_CUBE = 779
    SUBER_CUBE_IRR = 780
    REPORT_HYPER_CARD = 781
    METRIC = 1024
    SUBTOTAL_DEFINITION = 1025
    SYSTEM_SUBTOTAL = 1026
    METRIC_DMX = 1027
    METRIC_TRAINING = 1028
    STYLE = 1536
    METRIC_AGG = 1792
    FOLDER = 2048
    DEVICE = 2304
    PROMPT = 2560
    PROMPT_BOOLEAN = 2561
    PROMPT_LONG = 2562
    PROMPT_STRING = 2563
    PROMPT_DOUBLE = 2564
    PROMPT_DATE = 2565
    PROMPT_OBJECTS = 2566
    PROMPT_ELEMENTS = 2567
    PROMPT_EXPRESSION = 2568
    PROMPT_EXPRESSION_DRAFT = 2569
    PROMPT_DIMTY = 2570
    PROMPT_BIG_DECIMAL = 2571
    FUNCTION = 2816
    ATTRIBUTE = 3072
    ATTRIBUTE_ROLE = 3073
    ATTRIBUTE_TRANSFORMATION = 3074
    ATTRIBUTE_ABSTRACT = 3075
    ATTRIBUTE_RECURSIVE = 3076
    ATTRIBUTE_DERIVED = 3077
    ATTRIBUTE_SMART = 3078
    FACT = 3328
    DIMENSION_SYSTEM = 3584
    DIMENSION_USER = 3585
    DIMENSION_ORDERED = 3586
    DIMENSION_USER_HIERARCHY = 3587
    TABLE = 3840
    TABLE_PARTITION_MD = 3841
    TABLE_PARTITION_WH = 3842
    DATAMART_REPORT = 4096
    FACT_GROUP = 4352
    RESOLUTION = 4864
    ATTRIBUTE_FORM = 5376
    ATTRIBUTE_FORM_SYSTEM = 5377
    ATTRIBUTE_FORM_NORMAL = 5378
    SCHEMA = 5632
    FORMAT = 5888
    CATALOG = 6144
    CATALOG_DEFINITION = 6400
    COLUMN = 6656
    PROPERTY_GROUP = 6912
    PROPERTY_SET = 7168
    DB_ROLE = 7424
    DB_ROLE_DATA_IMPORT = 7425
    DB_ROLE_OAUTH = 7427
    DB_ROLE_GENERIC_DATA_CONNECTOR = 7430
    DB_LOGIN = 7680
    DB_CONNECTION = 7936
    PROJECT = 8192
    SERVER_DEFINITION = 8448
    USER = 8704
    USER_GROUP = 8705
    TRANSMITTER = 8705
    CONFIGURATION = 9216
    REQUEST = 9472
    SEARCH = 9984
    SEARCH_FOLDER = 10240
    DATAMART = 10496
    FUNCTION_PACKAGE_DEFINITION = 10752
    ROLE = 11008
    ROLE_TRANSFORMATION = 11009
    SECURITY_ROLE = 11264
    INBOX = 11520
    INBOX_MESSAGE = 11776
    CONSOLIDATION = 12032
    CONSOLIDATION_MANAGED = 12033
    CONSOLIDATION_ELEMENT = 12288
    SCHEDULE_EVENT = 12544
    SCHEDULE_OBJECT = 12800
    SCHEDULE_TRIGGER = 13056
    LINK = 13312
    DB_TABLE = 13568
    DB_TABLE_PMT = 13569
    TABLE_SOURCE = 13824
    DOCUMENT_DEFINITION = 14080
    REPORT_WRITING_DOCUMENT = 14081
    DOCUMENT_THEME = 14082
    DOCUMENT_BOT = 14084
    DRILL_MAP = 14336
    DBMS = 14592
    MD_SECURITY_FILTER = 14848
    MONITOR_PERFORMANCE = 15104
    MONITOR_JOBS = 15105
    MONITOR_USER_CONNECTIONS = 15106
    MONITOR_DB_CONNECTIONS = 15107
    PROMPT_ANSWER = 15232
    PROMPT_ANSWERS = 15360
    PROMPT_ANSWER_BOOLEAN = 15233
    PROMPT_ANSWER_LONG = 15234
    PROMPT_ANSWER_STRING = 15235
    PROMPT_ANSWER_DOUBLE = 15236
    PROMPT_ANSWER_DATE = 15237
    PROMPT_ANSWER_OBJECTS = 15238
    PROMPT_ANSWER_ELEMENTS = 15239
    PROMPT_ANSWER_EXPRESSION = 15240
    PROMPT_ANSWER_EXPRESSION_DRAFT = 15241
    PROMPT_ANSWER_DIMTY = 15242
    PROMPT_ANSWER_BIG_DECIMAL = 15243
    PROMPT_ANSWER_INT64 = 15244
    GRAPH_STYLE = 15616
    CHANGE_JOURNAL_SEARCH = 15872
    BLOB_UNKNOWN = 16128
    BLOB_OTHER = 16129
    BLOB_IMAGE = 16130
    BLOB_PROJECT_PACKAGE = 16131
    BLOB_EXCEL = 16132
    BLOB_HTML_TEMPLATE = 16133
    DASHBOARD_TEMPLATE = 16384
    FLAG = 16640
    CHANGE_JOURNAL = 16896
    EXTERNAL_SHORTCUT_UNKNOWN = 17152
    EXTERNAL_SHORTCUT_URL = 17153
    EXTERNAL_SHORTCUT_SNAPSHOT = 17154
    EXTERNAL_SHORTCUT_TARGET = 17408
    RECONCILIATION = 17664
    SYSTEM_PALETTE = 17920
    CUSTOM_PALETTE = 17921
    THRESHOLDS = 18432
    COMMAND_MANAGER_SCRIPT = 19456
    SCRIPT = 19457
    JUPYTER_NOTEBOOK_SCRIPT = 19458
    DATASOURCE_SCRIPT = 19459
    TRANSACTION_SCRIPT = 19460
    APPLICATION = 19968
    APPLICATION_POWERPOINT = 19969
    APPLICATION_EXCEL = 19970
    APPLICATION_TEAMS = 19971
    APPLICATION_TABLEAU = 19972
    TIMEZONE_SYSTEM = 20224
    TIMEZONE_CUSTOM = 20225
    CALENDAR_SYSTEM = 20736
    CALENDAR_CUSTOM = 20737
    INTERFACE_LANGUAGE_SYSTEM = 21760
    INTERFACE_LANGUAGE_CUSTOM = 21761
    WORKFLOW = 22016
    IAM_SECRET_VAULT = 21002
    NUGGETS_FILE = 23040
    STREAM_CHANNEL = 23552
    SUBSCRIPTION_ADDRESS = 65281
    SUBSCRIPTION_CONTACT = 65282
    SUBSCRIPTION_INSTANCE = 65283
    NOT_SUPPORTED = None

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value


class ExtendedType(Enum):
    RESERVED = 0
    RELATIONAL = 1
    MDX = 2
    CUSTOM_SQL_FREE_FORM = 3
    CUSTOM_SQL_WIZARD = 4
    DATA_IMPORT_GENERAL = 256
    DATA_IMPORT_FILE_EXCEL = 272
    DATA_IMPORT_FILE_TEXT = 288
    DATA_IMPORT_CUSTOM_SQL = 304
    DATA_IMPORT_TABLE = 320
    DATA_IMPORT_OAUTH = 336
    DATA_IMPORT_OAUTH_SFDC = 337
    DATA_IMPORT_OAUTH_GDRIVE = 338
    DATA_IMPORT_OAUTH_DROPBOX = 339
    _ = 341
    DATA_IMPORT_CUSTOM_SQL_WIZARD = 352
    DATA_IMPORT_SPARK = 416
    DATA_IMPORT_DATASET = 448
    DATA_IMPORT_OLAP_CUBE = 4096
    DATA_IMPORT_LIVE_CUBE = 4097

    def __new__(cls, value):
        member = object.__new__(cls)
        member._value_ = value
        return member

    def __int__(self):
        return self.value


TypeOrSubtype = (
    int | ObjectTypes | ObjectSubTypes | list[int | ObjectTypes | ObjectSubTypes]
)



import logging

from mstrio import config
from mstrio.connection import Connection
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import Entity
from mstrio.utils.helper import _prepare_objects as filter_objects
from mstrio.utils.helper import delete_none_values
from mstrio.utils.response_processors import drivers
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


def list_drivers(
    connection: Connection, to_dictionary: bool = False, **filters
) -> list['Driver'] | list[dict]:
    """Get all drivers as list of Driver objects or dictionaries.

    Optionally filter the drivers by specifying filters.

    Args:
        connection: Strategy One connection object
        to_dictionary: If True returns a list of Driver dicts,
           otherwise returns a list of Driver objects
       **filters: Available filter parameters:
           ['id', 'name', 'is_enabled', 'is_odbc']
    """
    return Driver.list(connection, to_dictionary, **filters)


@class_version_handler('11.3.0960')
class Driver(Entity):
    """Object representation of Strategy One Driver

    Attributes:
        connection: A Strategy One connection object
        id: Driver's ID
        name: Driver's name
        is_enabled: specifies if a Driver is enabled
        is_odbc: specifies if a Driver is of ODBC type
        description: Object description
        abbreviation: Object abbreviation
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: User object that is the owner
        icon_path: Object icon path
        view_media: View media settings
        ancestors: List of ancestor folders
        certified_info: Certification status, time of certification, and
            information about the certifier (currently only for document and
            report)
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.DRIVER
    _API_GETTERS = {
        (
            'name',
            'description',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'comments',
            'owner',
        ): objects_processors.get_info,
        ('id', 'name', 'is_enabled', 'is_odbc'): drivers.get,
    }
    _API_PATCH: dict = {
        'enabled': (drivers.update, 'patch'),
        (
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }

    def __init__(
        self, connection: Connection, id: str | None = None, name: str | None = None
    ):
        """Initialize Driver object by passing ID or name.
        When `id` is provided, `name` is omitted.

        Args:
            connection (Connection): Strategy One connection object
            id (str): ID of Driver
            name (str): name of Driver
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            result: list[dict] = self.list(connection, name=name, to_dictionary=True)

            if result:
                object_data = result[0]
                self._init_variables(**object_data, connection=connection)
            else:
                raise ValueError(f"There is no Driver named: '{name}'.")
        elif id == '':
            raise ValueError("Driver's ID cannot be an empty string.")
        else:
            super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.is_enabled = kwargs.get('is_enabled')
        self.is_odbc = kwargs.get('is_odbc')

    def alter(
        self,
        is_enabled: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ) -> None:
        """Update properties of a Driver

        Args:
            is_enabled (bool, optional): specifies if a Driver is enabled
            comments (str, optional): long description of the Driver
            owner: (str, User, optional): owner of the Driver
        """
        if isinstance(owner, User):
            owner = owner.id
        properties = {
            'enabled': is_enabled,
            'comments': comments,
            'owner': owner,
        }
        not_none_properties = delete_none_values(properties, recursion=False)
        self._alter_properties(**not_none_properties)

        if config.verbose:
            logger.info(f"Updated driver: '{self.name}' with ID: {self.id}.")

    @classmethod
    def list(
        cls, connection: Connection, to_dictionary: bool = False, **filters
    ) -> list[type['Driver']] | list[dict]:
        """Get all driver as list of Driver objects or dictionaries.

        Optionally filter the drivers by specifying filters.

        Args:
            connection (Connection): Strategy One connection object
            to_dictionary (bool): If True returns a list of Driver dicts,
               otherwise returns a list of Driver objects
           **filters: Available filter parameters:
               ['id', 'name', 'is_enabled', 'is_odbc']
        """

        objects = filter_objects(drivers.get_all(connection), filters)

        if to_dictionary:
            return objects

        return cls.bulk_from_dict(source_list=objects, connection=connection)

    def enable(self):
        """Enable driver."""
        if not self.is_enabled:
            self.alter(is_enabled=True)

    def disable(self):
        """Disable driver."""
        if self.is_enabled:
            self.alter(is_enabled=False)



"""mstrio: Simple and Secure Access to Strategy One Data

Mstrio provides a high-level interface for Python and is designed
to give data scientists, developers, and administrators simple and secure
access to their Strategy One environment.

It wraps Strategy One REST APIs into simple workflows, allowing users
to fetch data from cubes and reports, create new datasets, add new data
to existing datasets, and manage Users/User Groups, Servers, Projects,
and more. Since it enforces Strategy One’s user and object security model,
you don’t need to worry about setting up separate security rules.

With mstrio-py for data science, it’s easy to integrate cross-departmental,
trustworthy business data in machine learning workflows and enable
decision-makers to take action on predictive insights in Strategy One Reports,
Dashboards, HyperIntelligence Cards, and customized, embedded analytical
applications.

With mstrio-py for system administration, it’s easy to minimize costs by
automating critical, time-consuming administrative tasks, even enabling
administrators to leverage the power of Python to address complex
administrative workflows for maintaining a Strategy One environment.
"""

__title__ = "mstrio-py"
__version__ = "11.5.6.101"  # NOSONAR
__license__ = "DUAL-TIER LICENSE AGREEMENT"
__description__ = "Python interface for the Strategy One REST API"
__author__ = "Strategy One"
__author_email__ = "bkaczynski@strategy.com"

from .utils.dev_helpers import what_can_i_do_with  # noqa: F401
__build__ = "11.5.0600.00028"



import logging
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from mstrio import config

if TYPE_CHECKING:
    from mstrio.project_objects import Prompt

logger = logging.getLogger(__name__)


def get_prompt_answer(
    prompt: dict, prompt_answers: list[dict], force: bool
) -> Any | list[Any]:
    answer = prompt.get('answers')
    if not force and answer:
        input_msg = (
            f"Prompt: '{prompt.get('name')}' already has an answer. "
            "Would you like to keep the current answer? "
            "(Enter 'Y' to keep it, or 'N' to provide a new one from the "
            "`prompt_answers` list): "
        )
        if input(input_msg) != 'Y':
            answer = None

    if force or not answer:
        answer = next(
            (
                p.get('answers')
                for p in prompt_answers
                if p.get('key') == prompt.get('key')
            ),
            None,
        )
        if not answer:
            if not prompt.get('required'):
                return None

            msg = (
                "Cannot find an answer for the prompt: "
                f"`{prompt.get('name')}` in the `prompt_answers` list."
            )

            if force:
                raise ValueError(msg)

            input_msg = f"{msg} Please provide an answer:"
            answer = input(input_msg)

    return answer


def answer_prompts_helper(
    instance_id: str,
    prompt_answers: list['Prompt'],
    get_status_func: Callable,
    get_prompts_func: Callable,
    answer_prompts_func: Callable,
    force: bool,
) -> bool:
    """Helper function to answer prompts for a report or dashboard/document.

    Args:
        instance_id (str): Instance ID of the report or dashboard/document.
        prompt_answers (list[Prompt]): List of Prompt class objects answering
            the prompts.
        get_status_func (callable): Function to get the status of
            the report or dashboard/document.
        get_prompts_func (callable): Function to get the prompts for
            the report or dashboard/document.
        answer_prompts_func (callable): Function to answer the prompts for
            the report or dashboard/document.
        force (bool): If True, then the report/document/dashboard's existing
            prompt will be overwritten by ones from the prompt_answers list,
            and additional input from the user won't be asked.
            Otherwise, the user will be asked for input if the prompt
            is not answered, or if prompt was already answered.

    Returns:
        bool: True if prompts were answered successfully, False otherwise.
    """
    instance_status = get_status_func().json().get('status')

    if instance_status != 2:
        if config.verbose:
            msg = (
                'There are no unanswered prompts '
                f'for the instance with ID: {instance_id}'
            )
            logger.warning(msg)
        return False

    prompt_answers = [prompt.to_dict() for prompt in prompt_answers]

    while instance_status == 2:
        prompted_instance = get_prompts_func().json()

        body = [
            {
                'key': prompt.get('key'),
                'id': prompt.get('id'),
                'name': prompt.get('name'),
                'type': prompt.get('type'),
                'answers': answer,
            }
            for prompt in prompted_instance
            if (answer := get_prompt_answer(prompt, prompt_answers, force)) is not None
        ]

        answer_prompts_func(body={'prompts': body})

        instance_status = get_status_func().json().get('status')

    if config.verbose:
        msg = f'Prompts for the instance with ID: {instance_id} have been answered.'
        logger.info(msg)

    return True



"""Create an email subscription with prompts.
1. Connect to the environment using data from workstation
2. Create a list of prompts for the report
3. Answer the prompts
4. Create an email subscription with the answered prompts
"""

from mstrio.connection import get_connection
from mstrio.distribution_services import Content, EmailSubscription, Schedule
from mstrio.project_objects import Prompt, Report

PROJECT_NAME = 'MicroStrategy Tutorial'  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Prepare a list of answers for the prompts in the report
prompts = [
    Prompt(
        type='OBJECTS',
        answers=[
            {"name": "Cost", "id": "7FD5B69611D5AC76C000D98A4CC5F24F", "type": "metric"}
        ],
        key='8891A8AF4A747A3EA8506DBB6189F252@0@10',
        id='8891A8AF4A747A3EA8506DBB6189F252',
        name='Predefined list of objects',
        use_default=False,
    ),
    Prompt(
        type='EXPRESSION',
        answers={
            "content": "Item (ID) = 23",
            "expression": {
                "operator": "And",
                "operands": [
                    {
                        "operator": "Equals",
                        "operands": [
                            {
                                "type": "form",
                                "attribute": {
                                    "id": "8D679D4211D3E4981000E787EC6DE8A4",
                                    "name": "Item",
                                },
                                "form": {
                                    "id": "45C11FA478E745FEA08D781CEA190FE5",
                                    "name": "ID",
                                },
                            },
                            {
                                "type": "constant",
                                "dataType": "Integer",
                                "value": "23",
                            },
                        ],
                    }
                ],
            },
        },
        key='742FE7A7486EE982FC006C98D4048A1B@0@10',
        id='742FE7A7486EE982FC006C98D4048A1B',
        name='Attributes of the Products hierarchy',
        use_default=False,
    ),
    Prompt(
        type='ELEMENTS',
        answers=[{"id": "h11;8D679D4B11D3E4981000E787EC6DE8A4", "name": "Canada"}],
        key='0CF9922747C3C94DAA18F489C36532D2@0@10',
        id='0CF9922747C3C94DAA18F489C36532D2',
        name='Elements of Region',
        use_default=False,
    ),
    Prompt(
        type='ELEMENTS',
        answers=[
            {"id": "h1;8D679D3711D3E4981000E787EC6DE8A4", "name": "Books"},
            {"id": "h3;8D679D3711D3E4981000E787EC6DE8A4", "name": "Movies"},
        ],
        key='094AF11F4164B417491024AE00C2A122@0@10',
        id='094AF11F4164B417491024AE00C2A122',
        name='Elements of Category',
        use_default=False,
    ),
    Prompt(
        type='ELEMENTS',
        answers=[{"id": "h12;8D679D4F11D3E4981000E787EC6DE8A4", "name": "Business"}],
        key='17E83147401F9CEDAA07CF826DE4CFB4@0@10',
        id='17E83147401F9CEDAA07CF826DE4CFB4',
        name='Elements of Subcategory filtered by Elements of Category',
        use_default=False,
    ),
]

# Report with nested prompts
report = Report(connection=conn, name='6- Prompted Profit Trend')

report.answer_prompts(prompts)

# After answering the prompts, the report object will have the instance_id
# property set, which is required for creating a subscription with prompts
print(report.instance_id)

email_sub = EmailSubscription.create(
    connection=conn,
    name='<Name of the subscription>',
    project_name=PROJECT_NAME,
    contents=[
        Content(
            id=report.id,
            type=Content.Type.REPORT,
            personalization=Content.Properties(
                format_type=Content.Properties.FormatType.PDF,
                prompt=Content.Properties.Prompt(
                    enabled=True,
                    instance_id=report.instance_id,
                ),
            ),
        ),
    ],
    schedules=Schedule(conn, name='Monday Morning'),
    recipients=['54F3D26011D2896560009A8E67019608'],
    email_subject='<Email subject>',
)
